#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

"""
Classes and methods for laying the data location foundation for various types of CASA processing
"""
import logging
import os
import shutil

import sys
import tarfile
from distutils.dir_util import copy_tree
from pathlib import Path

from casa_envoy.interfaces import FoundationIF
from casa_envoy.schema import AbstractTextFile


class GeneralFoundation(FoundationIF):
    """
    Class to set up the general data foundation for anything running CASA.
    - Special attention to user provided file locations, must be placed in working directory.
    """

    def __init__(self, parameters: dict, metadata: AbstractTextFile):
        self.logger = logging.getLogger("casa_envoy")
        self.parent_path = parameters["parent_path"]
        self.working_dir = self.parent_path / "working"
        self.metadata = metadata

    def data_foundation(self):
        """
        All user provided files should be placed in the 'working' directory inorder for CASA to pick them up.

        :return:
        """
        self.logger.info("LAYING DATA FOUNDATION...")

        if "providedFiles" in self.metadata.content:
            provided_files = (
                self.metadata.content["providedFiles"].replace("[", "").replace("'", "").replace("]", "").split(",")
            )

            # move provided files from parent directory into working directory for CASA
            for file in provided_files:
                stripped_file = file.strip()
                if stripped_file:
                    shutil.copy(stripped_file, self.working_dir)

        self.logger.info("DATA FOUNDATION COMPLETE!")


class RestoreFoundation(FoundationIF):
    """
    Class to set up the data foundation for restores
    """

    def __init__(self, parameters: dict, metadata: AbstractTextFile):
        self.logger = logging.getLogger("casa_envoy")
        self.parameters = parameters
        self.parent_path = parameters["parent_path"]
        self.metadata = metadata

    def data_foundation(self):
        """
        CMS Restore requires two inputs: An EB and a Calibration for that EB
        After download, all data is in rawdata and the calibrations tables are contained
        in a tar file. We need to extract all the calibration tables to the products directory
        for CASA processing.

        :return:
        """

        self.logger.info("LAYING DATA FOUNDATION...")
        # ensure we're starting from the parent directory
        os.chdir(self.parent_path)
        self.extract_cal()
        self.set_permissions()
        self.logger.info("DATA FOUNDATION COMPLETE!")

    def extract_cal(self):
        self.logger.info("Extracting calibration tar file to products directory...")
        cal_name = self.metadata.content["fileSetIds"][1]
        cal_path = "./rawdata/" + cal_name + "/" + cal_name
        if Path(cal_path).exists():
            calibration = tarfile.open(cal_path)
            # extract to products
            calibration.extractall("./rawdata")
            calibration.close()
            src = "./rawdata/products"
            src2 = "./rawdata/" + cal_name[:-4] + "/products"
            if Path(src).exists():
                copy_tree(src, "./products/")
            elif Path(src2).exists():
                copy_tree(src2, "./products")
            else:
                self.logger.error(f"ERROR: calibration tar file {cal_name} failed to extract correctly!")

        else:
            self.logger.error(f"ERROR: calibration tar file {cal_name} not found in rawdata!")
            # abort since required data not found
            sys.exit(1)

    def set_permissions(self):
        self.logger.info("Ensuring correct file permissions....")
        path = Path("./products")
        for root, dirs, files in os.walk(path):
            for d in dirs:
                os.chmod(os.path.join(root, d), 0o755)
            for f in files:
                os.chmod(os.path.join(root, f), 0o755)
