#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

"""
Command-line script that wraps the functionality of `docker exec -it` to enter our workflow Docker container and
uses HTCondor's `condor_ssh_to_job` command to access a running workflow.
"""

import argparse
import http
import json
import logging
import os
import subprocess
import sys
from typing import Dict, List, Optional

import requests
from pycapo import CapoConfig

logger = logging.getLogger("wf_inspector")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

ClassAd = Dict

CLASS_AD_BATCH_ID = "JobBatchId"
CLASS_AD_CLUSTER_ID = "ClusterId"


class NoJobFoundError(ValueError):
    pass


class NoIdForWorkflowError(ValueError):
    pass


class NoSuchWorkflowRequestError(ValueError):
    pass


def make_arg_parser() -> argparse.ArgumentParser:
    """
    Make command-line argument parser with correct argument configuration

    :return: Argument parser
    """
    arg_parser = argparse.ArgumentParser(
        description="Inspect the working directory of a running workflow",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    arg_parser.add_argument(
        "workflow_request_id",
        type=int,
        action="store",
        help="the workflow request ID of the running workflow",
    )
    arg_parser.add_argument(
        "-c",
        "--condor",
        action="store_true",
        required=False,
        help="Return only the HTCondor ID for the given workflow request id",
    )
    return arg_parser


def is_docker() -> bool:
    """
    Determine if we're in a docker container
    :return: boolean
    """
    host = os.environ.get("HOSTNAME")

    if any(char.isdigit() for char in host):
        return True

    return False


def get_container_id_by_substr(substr: str) -> Optional[bytes]:
    """
    Get ID of running container searched on a substring of its name

    :param substr: Substring to search on
    :return: Container ID, if found; else None
    """
    if is_docker() is False:
        result = subprocess.run(["docker", "ps", "-q", "-f", f"name={substr}"], stdout=subprocess.PIPE, check=True)
        if result.returncode == 0:
            if len(result.stdout) > 0:
                # Container was found matching substr
                return result.stdout.strip()
            return None
        logger.error("Docker command not accessible. Docker is either not installed or you lack permissions to use it.")
        logger.info("Are you running this from within a Docker container? Docker is not installed in containers.")
    return None


def get_wf_request_htcondor_id(workflow_request_id: str) -> int:
    """
    Given a workflow request ID, get the HTCondor job ID that corresponds

    :param workflow_request_id: ID of workflow request
    :return: HTCondor job ID for workflow request
    :raises NoIdForWorkflow if no ID was found for the workflow
    """
    if os.environ["CAPO_PROFILE"] == "docker":
        workflow_url = "http://localhost:3456"
    else:
        # Non-local scenario; use normal CAPO setting
        workflow_url = CapoConfig().settings("edu.nrao.workspaces.WorkflowSettings").serviceUrl
    response = requests.get(
        # About "name" in URL: workflow name not necessary for this endpoint, so any string works here
        f"{workflow_url}/workflows/name/requests/{workflow_request_id}/htcondor_id"
    )
    if response.status_code == http.HTTPStatus.OK:
        htcondor_id = response.json()["htcondor_job_id"]

        if htcondor_id != "None":
            return int(htcondor_id)
        raise NoIdForWorkflowError

    raise NoSuchWorkflowRequestError(f"No workflow request found with ID {workflow_request_id}")


def get_running_class_ads(workflow_container_id: bytes) -> List[ClassAd]:
    """
    Query the cluster for a list of the ClassAds of all running jobs in JSON format

    :param workflow_container_id: ID of the running workflow container
    :return: List of ClassAd JSON
    """
    if workflow_container_id is None:
        running_jobs = subprocess.run(["condor_q", "-json"], stdout=subprocess.PIPE, check=True)
    else:
        running_jobs = subprocess.run(
            ["docker", "exec", "-it", workflow_container_id, "condor_q", "-json"], stdout=subprocess.PIPE, check=True
        )

    if running_jobs.stdout:
        running_jobs_json = json.loads(running_jobs.stdout)
        return running_jobs_json
    else:
        if workflow_container_id is None:
            workflow_container_id = "workspaces-workflow-1"
        raise NoJobFoundError(f"No running jobs found for container {workflow_container_id}.")


def is_dag_job(running_jobs_json: List[ClassAd], htcondor_job_id: int) -> bool:
    """
    Determine whether or not the given HTCondor job ID corresponds to a DAGMan job

    :param running_jobs_json: List of info about running jobs in JSON form
    :param htcondor_job_id: Given HTCondor ID
    :return: True if the job is a DAGMan job; otherwise False
    :raises NoJobFound if no job is found with the given htcondor_job_id
    """
    for job in running_jobs_json:
        if job.get(CLASS_AD_CLUSTER_ID) == htcondor_job_id:
            # Found specified job's JSON
            found_job = job

            # A job with this attribute is a DAG job
            return found_job.get(CLASS_AD_BATCH_ID) is not None
    raise NoJobFoundError(f"No running job found for the the ID {htcondor_job_id}.")


def get_dag_job_id(running_jobs_json: List[ClassAd], dagman_job_id: int) -> str:
    """
    Given the ID of a DAGMan job, find the currently running normal job associated with it

    NOTE: It is possible to have more than one DAG job running at once, but not within the current scope of the system.
      Therefore, this method will only look for a single running job.

    :param running_jobs_json: List of info about running jobs in JSON form
    :param dagman_job_id: Given DAGMan job ID
    :return: ID of running DAG job as a str
    :raises NoJobFound if no job is found associated with the DAG's ID
    """
    for job in running_jobs_json:
        batch_id_str = job.get(CLASS_AD_BATCH_ID)
        batch_id = int(float(batch_id_str)) if batch_id_str else None
        if batch_id == dagman_job_id:
            cluster_id = job.get(CLASS_AD_CLUSTER_ID)
            is_dag_child = batch_id and (cluster_id != batch_id)

            if is_dag_child:
                return cluster_id
    raise NoJobFoundError(f"No running job found for the the DAG with ID {dagman_job_id}.")


def ssh_to_condor_job(workflow_container_id: bytes, htcondor_job_id: int):
    """
    Execute a condor_ssh_to_job command inside the running workflow container to connect the user to their job

    :param workflow_container_id: ID of the running workflow container
    :param htcondor_job_id: ID of the HTCondor job that will be SSHed into
    """
    subprocess.run(
        ["docker", "exec", "-it", workflow_container_id, "condor_ssh_to_job", "-auto-retry", str(htcondor_job_id)],
        check=True,
    )


def list_running_jobs(workflow_container_id: bytes):
    """
    List currently running jobs on the cluster

    :param workflow_container_id: ID of the running workflow container
    """
    logger.warning("HTCondor job ID not found!")

    if workflow_container_id is None:
        running_jobs = subprocess.run(["condor_q"], stdout=subprocess.PIPE, check=True)
        condor_q_output = running_jobs.stdout.decode("ascii").strip()
        logger.info("Here are the jobs currently running on the cluster\n")
        logger.info("%s\n", condor_q_output)
        logger.info("Is your job still running?")
    else:
        running_jobs = subprocess.run(
            ["docker", "exec", "-it", workflow_container_id, "condor_q"], stdout=subprocess.PIPE, check=True
        )
        condor_q_output = running_jobs.stdout.decode("ascii").strip()
        logger.info("Here are the jobs currently running on the cluster\n")
        logger.info("%s\n", condor_q_output)
        logger.info("Is your job still running?")


def main():
    args = make_arg_parser().parse_args()
    request_id = args.workflow_request_id
    condor = args.condor

    workflow_container_id = get_container_id_by_substr("workflow")

    # Get HTCondor ID for workflow request
    logger.info("Getting HTCondor ID for workflow request %s", request_id)
    try:
        htcondor_job_id = get_wf_request_htcondor_id(request_id)
    except NoIdForWorkflowError:
        logger.warning("No HTCondor ID found for workflow request %s", request_id)
        list_running_jobs(workflow_container_id)
        return

    logger.info("Found HTCondor ID %s", htcondor_job_id)

    # Determine if the given job is part of a DAG
    running_jobs_json = get_running_class_ads(workflow_container_id)
    if is_dag_job(running_jobs_json, htcondor_job_id):
        logger.info("HTCondor job with ID %s is a DAG.", htcondor_job_id)
        logger.info("Locating running job for DAG...")
        try:
            htcondor_job_id = get_dag_job_id(running_jobs_json, htcondor_job_id)
        except NoJobFoundError:
            list_running_jobs(workflow_container_id)
            return

    if condor is True:
        logger.info("Success! Your job has ID %s.", htcondor_job_id)
    else:
        logger.info("Success! Your job has ID %s. Logging in now...", htcondor_job_id)
        ssh_to_condor_job(workflow_container_id, htcondor_job_id)
