import argparse

from pendulum import datetime

from ws_metrics.enums import FetchType
from ws_metrics.queries.constructor import QueryConstructor, QueryString
from ws_metrics.queries.query_definitions import CAPABILITY_EXECUTIONS, DOWNLOAD_VOLUME, OPERATIONS, OPERATIONS_COUNT


class LifeUniverseEverything:
    def __init__(self, connection, args: argparse.Namespace):
        self.conn = connection
        self.args = args
        self.constructor = QueryConstructor(self.conn)

    def get_total_cap_executions(self, capability_name: str) -> int:
        """
        Determine the total number of capability executions for a given capability that exist on this environment

        :param capability_name: the name of the capability to gather metrics for
        :return: the total number of capability executions
        """
        sql = QueryString(CAPABILITY_EXECUTIONS, False)
        params = {"capability_name": capability_name[0]}
        return self.constructor.construct_query(sql, params, FetchType.ONE)

    def get_total_executions_in_range(self, capability_name: str, beginning_date: datetime, end_date: datetime) -> int:
        """
        Determine the total number of executions of a specified capability

        :param capability_name: The name of the capability to gather metrics on
        :param beginning_date: The beginning of the date range of interest
        :param end_date: The end of the date range of interest
        :return: The number of capability executions that occurred in the specified date range
        """
        sql = QueryString(CAPABILITY_EXECUTIONS, True, 0)
        params = {"capability_name": capability_name, "beginning": beginning_date, "end": end_date}
        return self.constructor.construct_query(sql, params, FetchType.ONE)

    def get_request_data_volume(self, request_id: str) -> float:
        """
        Get the total downloaded data volume per capability request

        :param request_id: The capability request of interest
        :return: Data size in Kilobytes
        """
        sql = QueryString(DOWNLOAD_VOLUME, True, 0)
        params = {"request_id": request_id}
        return self.constructor.construct_query(sql, params, FetchType.ONE)

    def get_total_data_volume(self, beginning_date: datetime, end_date: datetime) -> float:
        """
        Determine the total data volume downloaded withing a specified date rge

        :param beginning_date: The beginning date of interest
        :param end_date: The ending date of interest
        :return: The total data size in Kilobytes
        """
        sql = QueryString(DOWNLOAD_VOLUME, True, 1)
        params = {"beginning": beginning_date, "end": end_date}
        return self.constructor.construct_query(sql, params, FetchType.ONE)

    def get_operations_report_info(self):
        """
        Run operations metrics report

        :return: report data
        """
        sql = QueryString(OPERATIONS, False)
        params = {}
        return self.constructor.construct_query(sql, params, FetchType.ALL)

    def get_operations_report_count(self, beginning_date: datetime, end_date: datetime) -> int:
        """
        Determine count of all datasets archived between two dates

        :return: Number of archive datasets
        """
        sql = QueryString(OPERATIONS_COUNT, False)
        params = {"beginning": beginning_date, "end": end_date}
        return self.constructor.construct_query(sql, params, FetchType.ALL)

    def run_metrics(self, args: argparse.Namespace):
        if args.capability:
            result = self.get_total_cap_executions(args.capability)
            print("\nTotal number of " + args.capability[0] + " capability executions: " + str(result[0]) + "\n")
        if args.between:
            result = self.get_total_executions_in_range(args.between[0], args.between[1], args.between[2])
            print(
                "\nTotal number of "
                + args.between[0]
                + " executions between "
                + args.between[1]
                + " and "
                + args.between[2]
                + ": "
                + str(result[0])
                + "\n"
            )
        if args.requestdatasize:
            result = self.get_request_data_volume(args.requestdatasize[0])
            if result is not None:
                print("\nDownload data size for request #" + args.requestdatasize[0] + " : " + str(result[0]) + "K")
            else:
                print("Request #" + args.requestdatasize[0] + " not found!")
        if args.datasize:
            result = self.get_total_data_volume(args.datasize[0], args.datasize[1])
            if result[0] is not None:
                print(
                    "\nTotal volume of downloaded data for range "
                    + args.datasize[0]
                    + " to "
                    + args.datasize[1]
                    + ": "
                    + str(result[0])
                    + "K"
                )
            else:
                print("Error: No download capability requests were found!")
